package org.openjdk.gcbench.retain;

import org.openjdk.jmh.annotations.*;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(1)
@BenchmarkMode(Mode.AverageTime)
@OutputTimeUnit(TimeUnit.NANOSECONDS)
@Threads(Threads.MAX)
@State(Scope.Benchmark)
public class Tree {

    Object retain;

    @Param({"100000"})
    private int size;

    @Setup
    public void setup() {
        List<Node> nodes = new ArrayList<>();
        for (int c = 0; c < size; c++) {
            nodes.add(new Node(null, null));
        }

        while (nodes.size() > 1) {
            List<Node> newNodes = new ArrayList<>();
            for (int c = 0; c < nodes.size(); c += 2) {
                if (c + 1 < nodes.size()) {
                    newNodes.add(new Node(nodes.get(c), nodes.get(c + 1)));
                } else {
                    // non-balanced tree
                    newNodes.add(new Node(nodes.get(c), null));
                }
            }
            nodes = newNodes;
        }

        retain = nodes.get(0);
    }

    @Benchmark
    public Object test() {
        // allocation pressure to trigger GCs
        return new Object();
    }

    private static class Node {
        final Node left;
        final Node right;
        final Object payload;
        public Node(Node left, Node right) {
            this.left = left;
            this.right = right;
            this.payload = new Object();
        }
    }

}
